#coding=utf-8

###################################################################
# Copyright 2013 Civa lin 林雪凡
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
###################################################################

"""
LoNote Bookshelf 書櫃模組

一本書並不足夠，此處提供了管理一大堆書本的介面。

===================================

類別說明：

    Bookshelf = 顧名思義是用來收藏一堆書的容器物件
"""

import collections.abc
import os.path
import glob
import json

import lonote.im3

class BookshelfError(Exception): pass # 基本錯誤

class Bookshelf(collections.abc.MutableMapping):
    '''管理書籍載入狀況的容器

        1. 可以管理同個資料夾下所有的書
            - 在檔案中記錄個別書籍是否要載入，以及書籍排序
            - 可以調整書籍排序。
            - 即使透過外在手段（如剪貼）直接移動檔案，依然能正確初始化
        2. 可以透過本物件索引到所有的 book
        '''

    cfg_file = 'configure.json'

    @property
    def __cfg_path(self):
        return os.path.join(self.__bookdir, type(self).cfg_file)

    def __book_path(self, bid):
        '''回傳 book 路徑'''
        return os.path.join(self.__bookdir,
                bid + lonote.im3.Book.filename_ext)

    def __init__(self, bookdir):
        '''管理指定目錄下所有書本'''
        self.__books = collections.OrderedDict()
        self.__bookdir = os.path.abspath(bookdir)

    def __bids_available(self):
        ''' 回傳 self.__bookdir 下所有可用的 bid '''
        match = os.path.join(self.__bookdir,
                '*' + lonote.im3.Book.filename_ext)
        bids = [ os.path.splitext( os.path.basename(path) )[0]
                for path in glob.glob(match)
                if os.path.isfile(path) ]
        return bids

    def __bids_reallyload(self):
        '''回傳真正要載入的書籍 pid 列表

            這涉及：
                1. 檔案是否存在
                2. cfg 是否要求它載入
                3. cfg 中的排列順序
            '''
        summary = self.bids_summary()
        return [ bid for bid, (shouldload, exists) in summary.items()
                if shouldload and exists ]

    def cfg(self):
        '''回傳現有的 cfg，或是初始化一個 cfg'''
        def cfg_default():
            '''回傳 cfg 的預設值，也就是預設載入全部 available 項目'''
            default = collections.OrderedDict(
                    [ (bid, True) for bid in self.__bids_available() ])
            return default
        # Real work
        try:
            with open(self.__cfg_path, mode = 'rt', encoding = 'utf8') as f:
                cfg = json.load(f,
                        object_pairs_hook = collections.OrderedDict)
        except (IOError, ValueError): # (讀不到, 內容無法解析)
            return cfg_default()
        else:
            return cfg

    def cfg_update(self, cfg = None):
        '''更新現存的 cfg

            cfg = None 表示由檔案中讀入
            '''
        def cfg_save(cfg):
            '''將 cfg 存入檔案中'''
            with open(self.__cfg_path, mode = 'wt', encoding = 'utf8') as f:
                json.dump(cfg, f, ensure_ascii = False, indent = 4)
        if cfg is None:
            cfg = self.cfg()
        available = self.__bids_available()
        new_items = [ (bid, True) for bid in available
                if bid in cfg ]
        cfg.update(new_items)
        cfg_save(cfg)

    def bids_summary(self):
        '''回傳 bids 的總結報告

            格式為 OrderedDict，如下：

                {bid: (cfg_shouldload, exists), ... }

                bid            = book id
                cfg_shouldload = cfg 中是否有要求載入？
                exists         = 現在是否存於檔案系統中？
            '''
        self.cfg_update()
        bids_avai = self.__bids_available()
        cfg = self.cfg()
        bids_incfg = cfg.keys()
        bids_all = [ bid for bid in bids_incfg ] + bids_avai
        summary = collections.OrderedDict()
        for bid in bids_all:
            summary[bid] = (
                    (bid in cfg.keys() and cfg[bid] == True) or (bid not in cfg.keys() and bid in bids_avai), # (cfg要求載入) or (cfg未指定但存在)
                    bid in bids_avai)
        return summary

    def books_load(self):
        '''依照順序，載入所有 __bids_shouldload() 回傳的書本'''
        self.books_close()
        for bid in self.__bids_reallyload():
            path = self.__book_path(bid)
            book = lonote.im3.Book(path, mode = 'load')
            self.__books[bid] = book

    def books_close(self):
        '''關閉所有書籍'''
        for b in self.__books.values():
            b.close()
        self.__books = collections.OrderedDict()

    def book_delete(self, bid):
        '''刪除一本書'''
        book = self.__books.pop(bid)
        book.flush()
        path = book.path
        bak_path = os.path.splitext(path)[0] + '.bak'
        if os.path.isfile(bak_path):
            os.remove(bak_path)
        book.close()
        os.rename(path, bak_path)
        self.books_load()

    @property
    def folder(self):
        return self.__bookdir

    # MutableMapping 操作

    def __getitem__(self, key):
        return self.__books[key]

    def __setitem__(self, key, value):
        info = 'You can set the Bookshelf directly, use "load" alternate.'
        raise BookshelfError(info)

    def __delitem__(self, key):
        del self.__books[key]

    def __iter__(self):
        return self.__books.__iter__()

    def __len__(self):
        return len(self.__books)

    def __contains__(self, item):
        return item in self.__books
