#coding=utf-8

###################################################################
# Copyright 2013 Civa lin 林雪凡
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
###################################################################

"""
LoNote intermediation convert 中介格式轉換模組

負責將其他資料格式，轉換為 LoNote 的標準資料格式 - IM。
目前，主要也就是將舊版的 LoNote 格式轉換為新版

=================================

參數命名法表記如下：

    src_url = 轉換的來源路徑
    dst_url = 轉換的目標路徑

"""

import lonote.im
from lonote.im3 import Book, TimeStandard, SafeId


def safeprint(*args, **kwargs):
    '''Avoid Unicode Error'''
    try:
        print(*args, **kwargs)
    except UnicodeEncodeError as e:
        print('## CANNOT DISPLAY IN YOUR TERMINAL ##')


class FileFormatUpdate:

    def from_2X_to_3X(src_url, dst_url, auto_ext=True):
        '''將 2.X 版的資料轉換成 3.X 版

            auto_ext 為真則自動確保 dst_url 處有正確加上副檔名
            '''
        if auto_ext and not dst_url.endswith(Book.filename_ext):
            dst_url = dst_url + Book.filename_ext

        im = lonote.im.IM()
        im.load(src_url)
        rootpid = im.uuid()
        with Book(dst_url, mode='create', rootpid=rootpid) as book:
            # 函式定義：單頁資料轉移
            def transform_paper(pid):
                '''將資料從 im 傳送到 book 中

                    book 內的 paper 需預先建立妥當
                    '''
                # 讀出資料
                title = im.title(pid)
                safeprint('Processing: {}'.format(title))
                content = im.content(pid)
                ctime = im.ctime(pid)
                mtime = im.mtime(pid)
                resources = []
                records = []
                for item in im.resource_list(pid):
                    item['data'] = im.resource(pid, name=item['name'])
                    resources.append(item)
                for item in im.record_list(pid):
                    item['content'] = im.record(pid, name=item['name'])
                    records.append(item)
                # 寫入資料
                paper = book.paper(pid)
                paper.name = title
                paper.sections.data_entity_update(
                    type(paper).default_sid, data=content)
                ctimetext = TimeStandard.totext(TimeStandard.asutc(ctime))
                mtimetext = TimeStandard.totext(TimeStandard.asutc(mtime))
                paper.sections.data_meta_update(
                    type(paper).default_sid,
                    ctime=ctimetext,
                    mtime=mtimetext)
                for res in resources:
                    paper.resources.data_create(
                        SafeId.unsafechars_remove(res['name']),
                        data=res['data'],
                        res_type=res['res_type'])
                for rec in records:
                    ftimetext = TimeStandard.totext(
                        TimeStandard.asutc(rec['ftime']))
                    paper.sections.data_create(
                        SafeId.unsafechars_remove(rec['name']),
                        data=rec['content'],
                        desc=rec['comment'],
                        ctime=ftimetext,
                        mtime=ftimetext)

            # 函式定義：遍歷
            def walk(pid):
                '''將指定 pid 頁下方的次級 paper 新建，並將資料轉移過去

                    指定的 pid 需已經存在
                    不變更指定的 pid paper 本身
                    '''
                for child_pid in im.children_list(pid):
                    book.paper_create(position=(pid, 0), new_pid=child_pid)
                    transform_paper(child_pid)
                    walk(child_pid)

            # 函式定義：轉移 extras
            def transform_toybox():
                '''將 extras 移動到 book.toybox 中'''
                # head + css 合併為 custom.head
                bookcss = im.extra('book.css')
                if bookcss:
                    bookcss = bookcss.replace('@charset "utf8";', '')  # dirty
                extrahead = im.extra('extra_head')
                head = '{extrahead}\n<style>\n{bookcss}\n</style>'.format(
                    extrahead=extrahead or '',
                    bookcss=bookcss or ''
                    )
                book.toybox.data_create(
                    'appended.head', data=head, toy_type='text')
                # wrapper
                wrapper = im.extra('insert_wrapper')
                if wrapper:
                    book.toybox.data_create(
                        'wrapper.json', data=wrapper, toy_type='json')
            # 實際執行
            transform_paper(rootpid)
            walk(rootpid)
            safeprint('Transform extra content...')
            transform_toybox()
            safeprint('Transform complete, flushing...')
            book.flush()
            im.close()
            safeprint('Convert complete!')
            # safeprint('\n## All Done! ## - New file in here:\n ==> {}'\
            #         .format(book.path))


class Export:
    pass


class Import:
    pass
