#coding=utf-8

###################################################################
# Copyright 2013 Civa lin 林雪凡
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
###################################################################

"""
LoNote server 伺服器模組

定義 LoNote 的伺服器 API，與 LoNote Client 或瀏覽器互動。
"""

import mimetypes
import json
import os
import urllib
import urllib.parse as UP
import urllib.request as UR
import webbrowser
import time
import threading
import logging
import collections
import atexit
import sys
import imghdr
import hashlib
import concurrent.futures as CF
import functools
import tempfile
import shutil
import pickle
import html.parser as HP
import configparser
import re




def workaround_bottle_import_error_in_win32():
    '''以下程式碼被用來在 Win32GUI 環境下繞過 bottle import 錯誤'''
    try:
        sys.stdout.write("")
        sys.stdout.flush()
    # TESTED: Should use AttributeError replace IOError in python 3.3
    except AttributeError:
        class dummyStream:
            ''' dummyStream behaves like a stream but does nothing. '''
            def __init__(self): pass
            def write(self,data): pass
            def read(self,data): pass
            def flush(self): pass
            def close(self): pass
        # and now redirect all default streams to this dummyStream:
        sys.stdout = dummyStream()
        sys.stderr = dummyStream()
        sys.stdin = dummyStream()
        sys.__stdout__ = dummyStream()
        sys.__stderr__ = dummyStream()
        sys.__stdin__ = dummyStream()

workaround_bottle_import_error_in_win32()
import bottle
# 傳送的最大正文大小（不含附件）為 1GB
bottle.BaseRequest.MEMFILE_MAX = 1024 * 1024 * 1024



import mako.template
import mako.lookup
from bs4 import BeautifulSoup
import diff_match_patch
import whoosh.index as WI
import whoosh.analysis as WA
import whoosh.fields as WF
import whoosh.qparser as WQ
import whoosh.highlight as WH
from whoosh.qparser.dateparse import DateParserPlugin
from bottle_compressor import CompressorPlugin

import lonote.info
import lonote.translate
from lonote.bookshelf import Bookshelf
from lonote.im3 import *
import lonote.imcv3

# 全域變數初始化
INDEX_DIR = None
USER_STATIC_DIR = None
CFG_DIR = None
CFG_LONOTE_DIR = None
bs = None
STATIC_DIR = None

LOGGING_FILE = None

RECENT_EDITED_FILE = None
RECENT_EDITED = None # deque of: (bid, pid, sid)

CONFIG_FILE = None
CONFIG = None # configparser

TRASHCAN_BID = '_trashcan'

INDEX_LOCK = threading.Lock()

TEMPLATE_LOOKUP = mako.lookup.TemplateLookup()
app = bottle.Bottle(catchall = True)


# ====================================
# 輔助程式碼
# ====================================

class _Helper:
    '''堆放伺服器用輔助函式的地方'''
    def str_recode(item):
        if type(item) is str:
            return item.encode('latin-1').decode('utf8')
        else:
            return item
    @classmethod
    def recode(cls, func):
        '''作為裝飾器，替所有字串類參數重新編解碼'''
        # bottlepy > 0.12.3 後不再需要，日後可望移除
        # @functools.wraps(func)
        # def wrapper(*args, **kwargs):
        #     args = ( str_recode(a) for a in args )
        #     kwargs = { key: cls.str_recode(kwargs[key]) for key in kwargs }
        #     return func(*args, **kwargs)
        # return wrapper
        return func

    def inv_recode(string):
        return string.encode('utf8').decode('latin-1')
    @lonote.translate.renew
    def tpl(tplname, **kwargs):
        '''幫忙簡化呼叫 template 的介面'''
        auth_code = bottle.request.get_cookie('auth_code')
        standard_info = {
                'bs': bs,
                'info': lonote.info,
                'a_type': get_auth_type(auth_code),
                'config': CONFIG}
        kwargs.update(standard_info)

        mtpl = TEMPLATE_LOOKUP.get_template(tplname)
        return mtpl.render(**kwargs)

    def compress(string):
        '''縮減傳送出去的空間'''
        return re.sub('[^\S\n]+', ' ', string, flags = re.MULTILINE)

    def recent_edited_add(item):
        '''將新項目加入到 recent_edited 中'''
        if item[0] == TRASHCAN_BID: # 忽略 TRASHCAN_BID
            return
        try:
            RECENT_EDITED.remove(item)
        except: pass
        RECENT_EDITED.appendleft(item)
        with open(RECENT_EDITED_FILE, 'wb') as f:
            pickle.dump(RECENT_EDITED, f, protocol = 3)

    @classmethod
    def recent_readed_add(cls, item):
        '''將新項目加入到 recent_readed 中'''
        if item[0] == TRASHCAN_BID: # 忽略 TRASHCAN_BID
            return
        recent_readed = cls.recent_readed()
        if recent_readed is None:
            recent_readed = collections.deque(maxlen = 10)
        try:
            recent_readed.remove(item)
        except: pass
        recent_readed.appendleft(item)
        bottle.response.set_cookie('recent_readed',
                value = recent_readed,
                secret = 'recent_readed',
                max_age = 60*60*24*365,
                path = '/')

    def recent_readed():
        '''取出 recent_readed'''
        recent_readed = bottle.request.get_cookie( 'recent_readed',
                secret = 'recent_readed',
                default = collections.deque(maxlen = 10))
        return recent_readed

# ====================================
# 搜尋與索引程式碼
# ====================================

class ToText(HP.HTMLParser):
    '''將Html轉換成純文字，並且合理地換行'''
    def __init__(self):
        super().__init__()
        self.invisible_tag = {'head', 'script', 'object', 'javascript', 'iframe', 'video', 'audio'}
        self.start_newline_tag = {'br', 'hr', 'pre', 'div','blockquote',
                'h1', 'h2', 'h3', 'h4', 'h5', 'h6'}
        self.end_newline_tag = {'pre', 'li', 'p', 'div', 'blockquote',
                'h1', 'h2', 'h3', 'h4', 'h5', 'h6'}
        self.indent_tag = {'ul', 'ol', 'blockquote'}
        # 資料儲藏點
        self.datas = []
        # 狀態訊息
        self.in_pre = 0
        self.in_invisible = 0
        self.in_indent_level = 0
    def handle_starttag(self, tag, attrs):
        # pre 開始偵測
        if tag == 'pre':
            self.in_pre += 1
        # 看不見開始偵測
        if tag in self.invisible_tag:
            self.in_invisible += 1
        if tag in self.indent_tag:
            self.in_indent_level += 1
        # 需換行偵測
        if tag in self.start_newline_tag:
            self.datas.append('\n')
    def handle_endtag(self, tag):
        # pre 閉合偵測
        if tag == 'pre' and self.in_pre >= 1:
            self.in_pre -= 1
        # 看不見閉合偵測
        if tag in self.invisible_tag and self.in_invisible >= 1:
            self.in_invisible -= 1
        if tag in self.indent_tag and self.in_indent_level >= 1:
            self.in_indent_level -= 1
        # 需換行偵測
        if tag in self.end_newline_tag:
            self.datas.append('\n')
    def handle_data(self, data):
        if self.in_invisible <= 0: # 要是看不見當然不用匯入
            if self.in_pre <= 0:
                data = data.replace('\n', '')
            if self.in_indent_level >= 1:
                self.datas.append(' ' * 4 * self.in_indent_level)
            self.datas.append(data)
    def reset(self):
        super().reset()
        self.datas = []
    def feed(self, data):
        data = HP.HTMLParser().unescape(data)
        super().feed(data)
        ans = ''.join(self.datas)
        self.reset()
        return ans

to_text_parser = ToText()

def html2txt(html):
    '''將 html 內文去掉標籤並轉為純文字

        html   = 含有 html 內容的字串
        return ==> 純文字字串
        '''
    return to_text_parser.feed(html)

# 分詞器：將中文字依單一字分詞，其他英文字依照一般 word 分詞
INDEX_ANALYZER = WA.RegexTokenizer( '([\u2E80-\u9FFF])|(\w+(\.?\w+)*)')\
        | WA.StopFilter( minsize = 0)\
        | WA.LowercaseFilter()
# index 格式
INDEX_SCHEMA = WF.Schema(
        path = WF.ID(unique = True, stored = True),
        bid = WF.ID(stored = True),
        pid = WF.ID(stored = True),
        sid = WF.ID(stored = True),
        pname = WF.TEXT(analyzer = INDEX_ANALYZER, stored = True),
        pdesc = WF.TEXT(analyzer = INDEX_ANALYZER, stored = True),
        sdesc = WF.TEXT(analyzer = INDEX_ANALYZER, stored = True),
        text = WF.TEXT(analyzer = INDEX_ANALYZER, stored = True, chars = True),
        mtime = WF.DATETIME(stored = True),
        ctime = WF.DATETIME(stored = True)
        )

def index_section(writer, bid, pid, sid):
    '''索引指定頁面'''
    path = '/edit/{bid}/{pid}/{sid}.html'.format(
            bid = bid, pid = pid, sid = sid)
    paper = bs[bid].paper(pid)
    html = paper.sections.data_entity(sid)
    text = html2txt(html)
    data_meta = paper.sections.data_meta(sid)
    mtime = TimeStandard.aslocal(TimeStandard.fromtext(data_meta['mtime']))
    ctime = TimeStandard.aslocal(TimeStandard.fromtext(data_meta['ctime']))
    sdesc = data_meta['desc']
    writer.update_document(
            path = path,
            bid = bid,
            pid = pid,
            sid = sid,
            pname = paper.name,
            pdesc = paper.desc,
            sdesc = sdesc,
            text = text,
            mtime = mtime,
            ctime = ctime)

def index_update_all():
    '''重新索引所有項目'''
    ix = WI.create_in(INDEX_DIR, schema = INDEX_SCHEMA)
    writer = ix.writer()
    for bid in bs.keys():
        if bid == TRASHCAN_BID: # Trashcan NOT need to index
            continue
        else:
            for pid in bs[bid].ids:
                for sid in bs[bid].paper(pid).sections.ids:
                    index_section(writer, bid, pid, sid)
    writer.commit()

def index_update_necessary():
    '''進行增量索引，如果索引不存在會自動重建'''
    if not INDEX_LOCK.acquire(blocking = False): # block?
        return
    def update_all_and_release():
        index_update_all()
        INDEX_LOCK.release() # Must Release!
    def clear_index():
        shutil.rmtree(INDEX_DIR)
        os.mkdir(INDEX_DIR)

    # check index mode
    try:
        if not WI.exists_in(INDEX_DIR):  # 索引不存在的話
            update_all_and_release()
            return
    except ValueError:  # 開啟時出現錯誤的話……
        clear_index()
        update_all_and_release()
        return

    ix = WI.open_dir(INDEX_DIR)
    if ix.schema != INDEX_SCHEMA:  # 索引格式與目前使用格式不合的話
        ix.close()
        clear_index()
        update_all_and_release()
        return
    # Incremential index
    # Real work: Find all already indexed item and her datas
    with ix.searcher() as searcher:
        indexed = {
            (fields['bid'], fields['pid'], fields['sid']):
                (str(fields['mtime']), fields['pname'], fields['pdesc'])
            for fields in searcher.all_stored_fields() }
    # Find all exist item
    existed = {} # {(bid, pid, sid): mtime, ...}
    for bid in bs.keys():
        if bid == TRASHCAN_BID: # Trashcan NOT need to index
            continue
        else:
            book = bs[bid]
            for pid in book.ids:
                paper = book.paper(pid)
                for sid in paper.sections.ids:
                    section_meta = paper.sections.data_meta(sid)
                    existed[(bid, pid, sid)] = (
                            str(TimeStandard.aslocal(TimeStandard.fromtext(
                                section_meta['mtime']))),
                            paper.name,
                            paper.desc)

    writer = ix.writer()
    # Delete: already indexed but no longer exists
    has_delete = False
    for bid, pid, sid in indexed.keys():
        if (bid, pid, sid) not in existed.keys():
            path = '/edit/{bid}/{pid}/{sid}.html'.format(
                    bid = bid, pid = pid, sid = sid)
            writer.delete_by_term('path', path)
            has_delete = True
    # Need Update & Add
    for bid, pid, sid in existed.keys():
        key = (bid, pid, sid)
        if key not in indexed.keys(): # Add
            index_section(writer, bid, pid, sid)
        elif indexed[key] != existed[key]: # update
            index_section(writer, bid, pid, sid)
    # END
    if has_delete:
        writer.commit(optimize = True) # rewrite index but slow
    else:
        writer.commit()
    INDEX_LOCK.release()

def index_search(key, pagenum = 1):
    '''基於當前索引進行搜尋

        key      = 使用者輸入的查詢字串
        pagenum = 取得查詢結果的哪個分頁
        回傳搜索結果的 data-search.html 頁
        '''
    # now = DT.datetime.now()
    # index_update_necessary()
    # print(DT.datetime.now() - now)
    ix = WI.open_dir(INDEX_DIR)
    with ix.searcher() as searcher:
        qp = WQ.QueryParser('text', INDEX_SCHEMA)
        qp.add_plugin(WQ.MultifieldPlugin(["text", "pname", "pdesc"]))
        qp.add_plugin(DateParserPlugin())
        query = qp.parse(key)
        resultspage = searcher.search_page(query,
                pagenum = pagenum, pagelen = 10,
                terms = True)
        resultspage.results.fragmenter = WH.PinpointFragmenter(
                autotrim = True)
        return _Helper.compress(_Helper.tpl('data-search.mako',
                resultspage = resultspage,
                key = key, doc_count = ix.doc_count()))

# ====================================
# auth 中間層
# ====================================

def get_auth_type(auth_code):
    '''藉由 auth_code 計算並回傳權限等級，輸入的 auth_code 為 md5 後的

        回傳值為字元串，含 v 表示有讀權限，含 e 表示編輯權限'''
    f_code = CONFIG['AUTHCODE']['full_access_code']
    f_code_enc = hashlib.md5(f_code.encode('utf-8')).hexdigest()
    v_code = CONFIG['AUTHCODE']['viewable_code']
    v_code_enc = hashlib.md5(v_code.encode('utf-8')).hexdigest()
    if not f_code or auth_code == f_code_enc:
        level = 'ev'
    elif not v_code or auth_code == v_code_enc:
        level = 'v'
    else:
        level = ''
    return level

class AuthMiddleware:
    '''提供 Middle ware 來進行基本使用者身份驗證'''
    def __init__(self, wsgi):
        self.wsgi = wsgi

    def __call__(self, environ, start_response):
        request = bottle.Request(environ)
        auth_code = request.get_cookie('auth_code')
        # 確認權限類型
        a_type = get_auth_type(auth_code)
        # 權限判斷
        if request.urlparts.path in {'/permission.html', '/func/index-update'} or request.urlparts.path.startswith('/static/'): # 無管制路徑不受保護
            return self.wsgi(environ, start_response)
        elif request.method == 'GET' and 'v' in a_type: # 讀權限檢查
            return self.wsgi(environ, start_response)
        elif request.method != 'GET' and 'e' in a_type: # 其他權限
            return self.wsgi(environ, start_response)
        else: # 拒絕存取
            if request.method == 'GET':
                start_response('302 Redirect', [('Location', '/permission.html'),])
                return []
            else:
                start_response('401 Unauthorized', [])
                return ["You don't have full access permission. Some operation (like save or delete) has be rejected!".encode('utf8')]

# ====================================
# process safe 中間層
# ====================================

class ProcessSafeMiddleware:
    '''當 process id 改變，則重新載入所有筆記本，確保快取安全'''
    @property
    def pid(self):
        pidfile = os.path.join(CFG_LONOTE_DIR, '.current_pid')
        with open(pidfile, mode = 'r', encoding = 'utf8') as f:
            return f.read()

    @pid.setter
    def pid(self, value):
        pidfile = os.path.join(CFG_LONOTE_DIR, '.current_pid')
        with open(pidfile, mode = 'w', encoding = 'utf8') as f:
            f.write(str(value))

    def __init__(self, wsgi):
        self.wsgi = wsgi
        try:
            current_pid = str(os.getpid()) # 不支援 mac
            self.pid = current_pid # update
        except: pass

    def __call__(self, environ, start_response):
        try: current_pid = str(os.getpid()) # 不支援 mac
        except: current_pid = '0'
        if self.pid != current_pid:
            bs.books_load()
            self.pid = current_pid # update
            logging.info('## RELOAD BOOKSHELF BY pid CHANGE! ##')

        return self.wsgi(environ, start_response)

# ====================================
# 伺服器工作 route
# ====================================

@app.error(404)
@lonote.translate.renew
def error404(error):
    # logging.warning('File Not Found: "{}"'.format(bottle.request.path))
    message = _('Nothing here, sorry...')
    return message

@app.route('/permission.html')
@lonote.translate.renew
def permission():
    return _Helper.compress(_Helper.tpl("permission.mako"))

# 靜態檔案

@app.route('/static/<filename:path>')
@_Helper.recode
def static(filename):
    '''讀取全站共用的靜態檔案'''
    rt = bottle.static_file(filename, root = USER_STATIC_DIR)
    if type(rt) == bottle.HTTPError:
        return bottle.static_file(filename, root = STATIC_DIR)
    else:
        return rt

@app.route('/<filename>')
@_Helper.recode
def siteroot(filename):
    '''回送 siteroot 下的檔案'''
    bottle.redirect('/static/siteroot/{}'.format(filename))

@app.route('/')
def root():
    bottle.redirect('/start.html')

@app.route('/start.html')
@lonote.translate.renew
def start():
    bottle.response.set_header('Cache-Control', 'no-store') # no-cache
    return _Helper.compress(_Helper.tpl('start.mako',
                recent_readed = _Helper.recent_readed(),
                recent_edited = RECENT_EDITED))

@app.route('/hotkey.html')
@lonote.translate.renew
def hotkey():
    return _Helper.compress(_Helper.tpl('hotkey.mako'))

@app.route('/bookshelf.html')
@lonote.translate.renew
def bookshelf():
    '''取得全部頁面列表'''
    bottle.response.set_header('Cache-Control', 'no-store') # no-cache
    return _Helper.compress(_Helper.tpl("bookshelf.mako"))

@app.route('/data-bookshelf.html')
@lonote.translate.renew
def data_bookshelf():
    '''取得全部頁面列表'''
    data = _Helper.compress(_Helper.tpl("data-bookshelf.mako"))
    cached_md5 = bottle.request.query.getunicode('cached_md5')
    if cached_md5:
        md5 = hashlib.md5(data.encode('utf-8')).hexdigest()
        if md5 == cached_md5:
            return '' # request browser side use the cache;
    return data

@app.route('/logging.txt')
def logging_file():
    '''回傳反序排列後的伺服器 logging 檔'''
    bottle.response.content_type = 'text/plain; charset=utf-8'
    with open(LOGGING_FILE, mode = "rt", encoding = "utf-8") as f:
        return [ line for line in reversed(f.readlines()) ]


# 編輯路徑

@app.route('/edit/<bid>/<pid>/<sid>.html')
@_Helper.recode
@lonote.translate.renew
def edit_article(bid, pid, sid):
    '''取得 content 頁'''
    try:
        book = bs.get(bid)
        paper = book.paper(pid)
        mtime = paper.sections.data_meta(sid)['mtime']
    except: raise bottle.HTTPError(404)
    # if mtime == bottle.request.headers.get('If-None-Match'): # no-change
    #     raise bottle.HTTPResponse('', 304)
    # bottle.response.set_header('etag', mtime)
    bottle.response.set_header('Cache-Control', 'no-store') # no-cache

    export = bottle.request.query.getunicode('export')
    if export == 'html':
        return _Helper.compress(_Helper.tpl("export_html.mako",
                book = book, paper = paper, sid = sid))
    else:
        _Helper.recent_readed_add( (bid, pid, sid) )
        return _Helper.tpl("edit.mako",
                book = book, paper = paper, sid = sid)

@app.put('/edit/<bid>/<pid>/<sid>.html')
@_Helper.recode
@lonote.translate.renew
def edit_article_put(bid, pid, sid):
    '''更新內容'''
    paper = bs[bid].paper(pid)
    paper_name = bottle.request.forms.getunicode('paper_name')
    paper_desc = bottle.request.forms.getunicode('paper_desc')
    section_desc = bottle.request.forms.getunicode('section_desc')
    patch_text = bottle.request.forms.getunicode('patch_text')
    new_section = bottle.request.forms.getunicode('new_section')
    content = bottle.request.forms.getunicode('content')
    basetime = bottle.request.forms.getunicode('basetime')
    basemd5 = bottle.request.forms.getunicode('basemd5')
    mtime = basetime # 初始假設不改變
    # 修改 paper 本身
    if paper_name is not None:
        paper.name = paper_name
    if paper_desc is not None:
        paper.desc = paper_desc
    # 新建 section
    if new_section is not None:
        try:
            paper.sections.data_create(sid, data = new_section)
            logging.info('Section created: "{}" in "{}" in "{}"'.format(
                sid, pid, bid))
            return
        except:
            info = _('Section already exists')
            raise bottle.HTTPError(400, info) # 企圖新建但偏偏存在
    # 以下均為對現存 Section 的操作
    def no_edit_conflict_assert():
        try: # 檢查 Section 存在
            ori_mtime = paper.sections.data_meta(sid)['mtime']
        except IMNotExist:
            info = _('Section not exists')
            raise bottle.HTTPError(400, info)
        if ori_mtime != basetime: # 檢查發來的新內容是基於哪一個更動時間
            info = _('Editing conflict: Your edition base on "{basetime}" version, but the server version is "{ori_mtime}". So your edition was refused (Not save) !').format(basetime = basetime, ori_mtime = ori_mtime)
            raise bottle.HTTPResponse(info, 400)
    # 差分儲存
    if all((patch_text, basemd5)):
        no_edit_conflict_assert()
        # 檢查雙方差分前原始資料是否相同
        old_content = paper.sections.data_entity(sid)
        old_md5 = hashlib.md5(old_content.encode('utf-8')).hexdigest()
        if basemd5 != old_md5:
            info = _('MD5 not match: Delta not base on the same content.')
            raise bottle.HTTPResponse(info, 403)
        # 透過 delta 復原新資料
        dmp = diff_match_patch.diff_match_patch()
        patch = dmp.patch_fromText(patch_text)
        patched_content, results = dmp.patch_apply(patch, old_content)
        # 寫入
        paper.sections.data_entity_update(sid,
                data = patched_content, create = False)
        mtime = TimeStandard.utcnowtext()
        paper.sections.data_meta_update(sid, mtime = mtime)
        _Helper.recent_edited_add( (bid, pid, sid) )
    # 完整儲存
    elif content is not None:
        no_edit_conflict_assert()
        paper.sections.data_entity_update(sid,
                data = content, create = False)
        mtime = TimeStandard.utcnowtext()
        paper.sections.data_meta_update(sid, mtime = mtime)
        _Helper.recent_edited_add( (bid, pid, sid) )
    # section_desc
    if section_desc is not None:
        no_edit_conflict_assert()
        paper.sections.data_meta_update(sid, desc = section_desc)
        mtime = TimeStandard.utcnowtext()
        paper.sections.data_meta_update(sid, mtime = mtime)
        _Helper.recent_edited_add( (bid, pid, sid) )
    logging.info('Section modified: "/edit/{bid}/{pid}/{sid}.html" (Paper: "{pname}")'.format(bid = bid, pid = pid, sid = sid, pname = paper.name))
    return mtime

@app.delete('/edit/<bid>/<pid>/<sid>.html')
@_Helper.recode
def edit_section_delete(bid, pid, sid):
    '''刪除操作 section'''
    try:
        book = bs.get(bid)
        paper = book.paper(pid)
        paper.sections.data_meta(sid)
    except IMError: raise bottle.HTTPError(404)
    if sid == type(paper).default_sid:
        info = _('Cannot remove the default section!')
        raise bottle.HTTPResponse(info, 400)
    else:
        paper.sections.data_delete(sid)
        logging.info('Section deleted: "/edit/{bid}/{pid}/{sid}.html" (Paper: "{pname}")'.format(bid = bid, pid = pid, sid = sid, pname = paper.name))
        return '/edit/{bid}/{pid}/{sid}.html'.format(
                bid = bid, pid = pid, sid = Paper.default_sid)

@app.delete('/edit/<bid>/<pid>')
@_Helper.recode
@lonote.translate.renew
def edit_article_delete(bid, pid):
    '''刪除操作 paper'''
    try: # 確認可以刪除
        book = bs.get(bid)
        paper = book.paper(pid)
        pname = paper.name
    except IMError: raise bottle.HTTPError(404)
    if bid != TRASHCAN_BID: # 現在要刪的東西不是垃圾桶中的東西
        try: # 檢查垃圾桶的存在超越性
            trashcan_book = bs.get(TRASHCAN_BID)
            trashcan_book.rootpid
        except AttributeError: # 垃圾桶不存在則新建
            trashcan_filename = TRASHCAN_BID + Book.filename_ext
            trashcan_path = os.path.join(bs.folder, trashcan_filename)
            with Book(trashcan_path, mode = 'create'): pass
            bs.books_load()
            trashcan_book = bs.get(TRASHCAN_BID)
            trashcan_book.name = _('※ Recycle Bin ※')
            # bs.loads 後要重新讀入
            book = bs.get(bid)
            paper = book.paper(pid)
        # 將東西複製到垃圾桶中作為備份
        src_url = os.path.join(bs.folder, book.filename)
        position = (trashcan_book.rootpid, 0)
        trashcan_book.book_import(
                src_url = src_url, src_pid = pid,
                dst_position = position,
                avoid_conflict = True) # 允許變更 pid

    if book.rootpid != pid: # 刪除紙張
        book.paper_delete(pid)
        logging.info('Paper deleted: "{pid}" (Paper: "{pname}") in book "{bid}"'.format(pid = pid, pname = pname, bid = bid))
    else: # 刪除書本
        bs.book_delete(bid)
        logging.info('Book deleted: "{}"'.format(bid))

@app.post('/edit/<bid>/<pid>/<sid>.html')
@_Helper.recode
@lonote.translate.renew
def edit_article_post(bid, pid, sid):
    '''新建'''
    try:
        book = bs.get(bid)
        paper = book.paper(pid)
    except IMError: raise bottle.HTTPError(400)
    new_child = bottle.request.forms.getunicode('new_child')
    new_append = bottle.request.forms.getunicode('new_append')
    if new_child is not None:
        position = (paper.pid, 0)
    if new_append is not None:
        parent_pid = book.pid_parent(paper.pid)
        order = book.pid_children(parent_pid).index(paper.pid) + 2
        # +2 是 base 0 index -> base 1 index 和 插入「後面」兩要素造成的
        position = (parent_pid, order)
    new_paper = book.paper_create(position)
    parent_pid = book.pid_parent(new_paper.pid)
    parent_paper = book.paper(parent_pid)
    new_paper.name = _('Unnamed Paper ← {pname}').format(
            pname = parent_paper.name)
    logging.info('Paper created: "{}" in "{}"'.format(new_paper.pid, bid))
    return '/edit/{bid}/{pid}/{sid}.html'.format(
            bid = bid, pid = new_paper.pid, sid = Paper.default_sid)


# 資源路徑

@app.route('/edit/<bid>/<pid>/<rid>')
@_Helper.recode
def edit_resource(bid, pid, rid):
    '''取得 resource'''
    try: # Check Resource Exist
        book = bs.get(bid)
        paper = book.paper(pid)
        res_meta = paper.resources.data_meta(rid)
    except: raise bottle.HTTPError(404)
    # Cache?
    mtime = TimeStandard.fromtext(res_meta.get('mtime'))
    lm = DT.datetime.strftime(mtime, "%a, %d %b %Y %H:%M:%S GMT")
    bottle.response.set_header('Last-Modified', lm)
    ims = bottle.request.environ.get('HTTP_IF_MODIFIED_SINCE')
    if ims:
        ims = ims.split(";")[0].strip()
    if ims and ims == lm: # Use Cache
        return bottle.HTTPResponse(status = 304)
    else: # No Cache
        bottle.response.content_type = mimetypes.guess_type(rid)[0]
        return paper.resources.data_entity(rid)

@app.delete('/edit/<bid>/<pid>/<rid>')
@_Helper.recode
def edit_resource_delete(bid, pid, rid):
    '''刪掉 resource'''
    try:
        book = bs.get(bid)
        paper = book.paper(pid)
        paper.resources.data_delete(rid)
        logging.info('Resource deleted: "{rid}" deleted in "{bid}" -> "{pid}" (Paper: "{pname}")'.format(rid = rid, bid = bid, pid = pid, pname = paper.name))
    except IMError: raise bottle.HTTPError(404)

@app.put('/edit/<bid>/<pid>/<rid>')
@_Helper.recode
def edit_resource_put(bid, pid, rid):
    '''更新 resource'''
    try:
        book = bs.get(bid)
        paper = book.paper(pid)
        data = bottle.request.body.read()
        res_type = bottle.request.query.getunicode('res_type')
        if res_type is None:
            bottle.HTTPError(400)
        try:
            paper.resources.data_create(rid, data = data,
                    res_type = res_type)
            logging.info('Resource created: "{rid}" created in "{bid}" -> "{pid}" (Paper: "{pname}")'.format(rid = rid, bid = bid, pid = pid, pname = paper.name))
        except IMAlreadyExist:
            paper.resources.data_entity_update(rid, data = data,
                    res_type = res_type)
            logging.info('Resource updated: "{rid}" updated in "{bid}" -> "{pid}" (Paper: "{pname}")'.format(rid = rid, bid = bid, pid = pid, pname = paper.name))
        except IMValueError:
            raise bottle.HTTPResponse('Filename unacceptable!', 403)
    except IMError: raise bottle.HTTPError(404)

# 獨立功能路徑

@app.route('/func/paperinfo/<bid>/<pid>')
@_Helper.recode
@lonote.translate.renew
def paperinfo(bid, pid):
    '''回傳 paper 的相關訊息，用在刪除前確認等地方'''
    try:
        book = bs.get(bid)
        paper = book.paper(pid)
        parent_pid = book.pid_parent(pid)
        if parent_pid:
            parent_name = book.paper(parent_pid).name
        else:
            parent_name = _('No exists, It\'s top paper.')
        children_name = [ book.paper(child_pid).name
                for child_pid in book.pid_children(pid) ]
        children_num = len(children_name)
        children_name = '\n'.join(
                [ '\t{}'.format(n) for n in children_name ])
        return _('This Paper is: "{pname}"\n\nWhich Book: "{bname}"\nParent Paper: "{parent_name}"\n\nHas Children: {children_num}\n{children_name}').format(bname = book.name, pname = paper.name, parent_name = parent_name, children_num = children_num, children_name = children_name)
    except IMError: raise bottle.HTTPError(404)

@app.post('/func/clearunused')
def clearunused():
    '''將指定頁面中所有未使用的 emb 類 resources 刪除'''
    bid = bottle.request.forms.getunicode('bid')
    pid = bottle.request.forms.getunicode('pid')
    book = bs.get(bid)
    paper = book.paper(pid)
    emb_rids = [ rid for rid in paper.resources.ids
            if paper.resources.data_meta(rid).get('res_type') == 'emb' ]
    contents = [ paper.sections.data_entity(sid)
            for sid in paper.sections.ids ]
    contents = [ c.lower() for c in contents ] # 小寫化
    contents = [ UP.unquote(c) for c in contents ] # 去除 % 編碼
    need_remove = emb_rids # 計算需移除的有哪些
    for content in contents:
        need_remove = { rid for rid in need_remove
                if (rid.lower() not in content) }
    # 移除
    for rid in need_remove:
        paper.resources.data_delete(rid)
        logging.info('Resource deleted: "{rid}" deleted in "{bid}" -> "{pid}"'.format(rid = rid, bid = bid, pid = pid))

@app.post('/func/imgemb')
def imgemb():
    '''將指定 section 的所有外連圖片轉為 resources 並改為內連'''
    def gen_filename(link):
        '''透過連結產生檔名，但不包含副檔名'''
        name = UP.urlparse(link).path.split('/')[-1] # 檔名全名
        # 剪裁副檔名
        try:
            name = name[:name.rindex('.')]
        except ValueError:
            pass # 本來就沒有副檔名
        # 加入 hashcode 儘量避免重複，但不要過長
        hashcode = hashlib.md5(link.encode('utf-8')).hexdigest()
        if name == '': # 預設名稱是空字串
            name = hashcode[:7]
        else:
            name = name + '_' + hashcode[:4]
        return name

    def download(url):
        '''return (data, url)'''
        try:
            data = UR.urlopen(url, timeout = 15).read()
            return (data, url)
        except:
            return (None, url)

    bid = bottle.request.forms.getunicode('bid')
    pid = bottle.request.forms.getunicode('pid')
    sid = bottle.request.forms.getunicode('sid')
    book = bs.get(bid)
    paper = book.paper(pid)
    content = paper.sections.data_entity(sid)
    content = UP.unquote(content) # 去除 % 編碼
    soups = { xid: BeautifulSoup(
                UP.unquote(paper.sections.data_entity(xid)))
            for xid in paper.sections.ids }
    # Real work: Parse external url
    soup = BeautifulSoup(content)
    imgs_soup = soup('img')
    links = [ str(img.get('src')) for img in imgs_soup ] # 轉成連結表
    request_links = { link for link in links   # 採用絕對路徑的連結
            if UP.urlparse(link).scheme in ('http', 'https') }
    link_names = {} # 替聯結產生檔名
    for link in request_links:
        link_names[link] = gen_filename(link)
    # 抓取並存入外部資源
    with CF.ThreadPoolExecutor(max_workers = 6) as executor:
        futures = [ executor.submit(download, url)
                for url, name in link_names.items() ]
        for future in futures:
            data, url = future.result()
            if data == None:
                continue
            rid = link_names.get(url)
            # 加上副檔名
            img_type = imghdr.what('', h = data)
            if img_type:
                rid = rid + '.' + img_type
            # 插入
            try:
                paper.resources.data_create(rid, data = data,
                        res_type = 'emb')
            except: pass
            else:
                for xid in soups:
                    for tag in soups[xid]('img', src = url):
                        tag['src'] = rid
                logging.info('External resource embedded: "{url}" -> "{rid}" now added to "{bid}" -> "{pid}"'.format(url = url, rid = rid, bid = bid, pid = pid))
    # 回存原始檔
    for xid in soups:
        paper.sections.data_entity_update(xid,
                data = ''.join([ str(s) for s in soups[xid].contents] ))

@app.post('/func/movepaper')
def movepaper():
    '''頁面移動'''
    moved_bid = bottle.request.forms.getunicode('moved_bid')
    moved_pid = bottle.request.forms.getunicode('moved_pid')
    parent_bid = bottle.request.forms.getunicode('parent_bid')
    parent_pid = bottle.request.forms.getunicode('parent_pid')
    order = int(bottle.request.forms.getunicode('order'))
    parent_book = bs.get(parent_bid)
    moved_book = bs.get(moved_bid)
    position = (parent_pid, order)
    # Real work
    if moved_book is parent_book:
        parent_book.paper_move(moved_pid, position = position)
        logging.info('Paper moved (in the same book): "{moved_bid}" -> "{moved_pid}" be moved'.format( moved_bid = moved_bid, moved_pid = moved_pid))
    else: # 跨筆記本移動
        src_url = os.path.join(bs.folder, moved_book.filename)
        parent_book.book_import(
                src_url = src_url, src_pid = moved_pid,
                dst_position = position,
                avoid_conflict = False)
        # 刪除原始殘留物
        if moved_book.rootpid == moved_pid:
            bs.book_delete(moved_bid)
            logging.info('Book moved in to the other book: "{moved_bid}" are moved in to the book -> "{parent_bid}"'.format( moved_bid = moved_bid, parent_bid = parent_bid))
        else:
            moved_book.paper_delete(moved_pid)
            logging.info('Paper moved (cross books): "{moved_bid}" -> "{moved_pid}" are moved in to the book -> "{parent_bid}"'.format( moved_bid = moved_bid, moved_pid = moved_pid, parent_bid = parent_bid))

@lonote.translate.renew
@app.post('/func/book-export-ready')
def book_export_ready():
    '''將子書項目移動為獨立書籍，並自動載入到 bs 中'''
    moved_bid = bottle.request.forms.getunicode('moved_bid')
    moved_pid = bottle.request.forms.getunicode('moved_pid')
    bids_list = json.loads(bottle.request.forms.getunicode('bids_list'))
    new_bid = bottle.request.forms.getunicode('new_bid')
    if new_bid == '':
        raise bottle.HTTPResponse(
                _('Empty filename are not acceptable! cancel.'), 400)
    if not new_bid.endswith(Book.filename_ext):
        new_bid = new_bid + Book.filename_ext
    new_path = os.path.join(bs.folder, new_bid)
    moved_book = bs.get(moved_bid)
    # Real work
    try:
        moved_book.book_export(url = new_path, pid = moved_pid)
    except IMError as e:
        raise bottle.HTTPResponse(str(e), 400)
    else:
        logging.info('Paper hive off: "{moved_pid}" in "{moved_bid}" now become a individual book call "{new_bid}"'.format(moved_pid = moved_pid, moved_bid = moved_bid, new_bid = new_bid))
        moved_book.paper_delete(moved_pid)
        # 用 new_bid 替換空字串，然後重新排序
        bids_list = [ bid or new_bid for bid in bids_list ]
        bs_cfg_update(bids_list)

@app.post('/func/book-re-order')
def book_re_order():
    '''將書架中的書籍順序重排'''
    bids_list = json.loads(bottle.request.forms.getunicode('bids_list'))
    bs_cfg_update(bids_list)
    logging.info('Books re-order')

def bs_cfg_update(bids_list):
    '''重排書籍順序

        bids_list = [ bid1, bid2, bid3, ... ]
                    新順序
        '''
    new_cfg = collections.OrderedDict(
            [ (bids, True) for bids in bids_list])
    ori_cfg = bs.cfg()
    for ori_key in ori_cfg.keys():
        if ori_key not in new_cfg:
            new_cfg[ori_key] = False
    bs.cfg_update( cfg = new_cfg )
    bs.books_load()

@app.route('/func/download/<bid>/<fname>')
@_Helper.recode
def download_book(bid, fname):
    '''下載書籍'''
    book = bs.get(bid)
    if book is None: # download nonexist book
        raise bottle.HTTPResponse(_('Book are not exists!'), 400)
    book.flush()
    bottle.response.content_type = 'application/lo3'
    bottle.response.set_header('Content-Disposition', 'attachment')
    with open(book.path, mode = "rb") as f:
        logging.info('Book Downloaded: "{}"'.format(bid))
        return f.read()

@app.route('/func/download/<bid>/<pid>/<name>')
@_Helper.recode
def download_subbook(bid, pid, name):
    '''下載匯出的書籍'''
    try: # make sure exists
        book = bs.get(bid)
        paper = book.paper(pid)
    except:
        raise bottle.HTTPResponse(_('Book or paper are not exists!'),
                400)
    with tempfile.TemporaryDirectory(prefix = 'lonote_') as temp_dir:
        temp_file = os.path.join(temp_dir, 'tempbook' + Book.filename_ext)
        book.book_export(url = temp_file, pid = pid)
        with Book(url = temp_file, mode = 'load') as new_book:
            new_book.flush()
        with open(temp_file, mode = 'rb') as f:
            bottle.response.content_type = 'application/lo3'
            bottle.response.set_header('Content-Disposition', 'attachment')
            logging.info('Sub-book exported: "{name}" from "{bid}".'.format(
                name = name, bid = bid))
            return f.read()

@app.post('/func/createbook')
@lonote.translate.renew
def create_book():
    '''新建全新筆記簿'''
    new_bid = bottle.request.forms.getunicode('new_bid')
    if not new_bid.endswith(Book.filename_ext):
        new_filename = new_bid + Book.filename_ext
    new_path = os.path.join(bs.folder, new_filename)
    try:
        with Book(new_path, mode = 'create'): pass
    except:
        raise bottle.HTTPResponse(_('This filename are not acceptable, please try another.'), 400)
    else:
        bs.books_load()
        book = bs.get(new_bid)
        book.name = os.path.splitext(new_filename)[0]
        link = '/edit/{bid}/{pid}/{sid}.html'.format(
                bid = new_bid, pid = book.rootpid, sid = Paper.default_sid)
        link = _Helper.inv_recode(link)
        logging.info('Book created: "{}"'.format(new_bid))
        bottle.redirect(link)

@app.post('/func/uploadbook')
@lonote.translate.renew
def upload_book():
    '''上傳筆記簿'''
    def file_not_acceptable():
        '''raise 不接受此一檔案的訊息'''
        raise bottle.HTTPResponse(_('This file are not acceptable!'), 400)
    # def redirect(bid):
    #     '''重定向到新筆記本的首頁'''
    #     book = bs.get(bid)
    #     link = '/edit/{bid}/{pid}/{sid}.html'.format(
    #             bid = bid, pid = book.rootpid, sid = Paper.default_sid)
    #     link = _Helper.inv_recode(link)
    #     bottle.redirect(link)
    for fileupload in bottle.request.POST.getall('uploadfile'):
        filename = fileupload.raw_filename
        fname_ext = os.path.splitext(filename)[1]
        if fname_ext == Book.filename_ext:
            new_path = os.path.join(bs.folder, filename)
            if os.path.exists(new_path):
                raise bottle.HTTPResponse(_('This filename are already exists! If you really want to upload this book, please rename it and upload again.'), 400)
            # Real work
            with tempfile.TemporaryDirectory(prefix = 'lonote_') as temp_dir:
                temp_path = os.path.join(temp_dir, filename)
                with open(temp_path, mode = "bx") as f:
                    f.write(fileupload.file.read())
                    fileupload.file.close()
                try: # try to load
                    with Book(temp_path, mode = 'load'): pass
                except:
                    file_not_acceptable()
                else:
                    shutil.move(temp_path, new_path)
                    logging.info('Book uploaded: "{}"'.format(filename))
                    new_bid = os.path.splitext(filename)[0]
                    bs.books_load()
        elif fname_ext == '.lobook': # 轉檔再載入
            new_filename = os.path.splitext(filename)[0] + Book.filename_ext
            new_bid = os.path.splitext(filename)[0]
            new_path = os.path.join(bs.folder, new_filename)
            if os.path.exists(new_path):
                raise bottle.HTTPResponse(_('This filename are already exists! If you really want to upload this book, please rename it and upload again.'), 400)
            with tempfile.TemporaryDirectory(prefix = 'lonote_') as temp_dir:
                temp_src_path = os.path.join(temp_dir, filename)
                temp_dst_path = os.path.join(temp_dir, new_filename)
                with open(temp_src_path, mode = "bx") as f: # 先寫入
                    f.write(fileupload.file.read())
                    fileupload.file.close()
                lonote.imcv3.FileFormatUpdate.from_2X_to_3X(
                        src_url = temp_src_path, dst_url = temp_dst_path)
                shutil.move(temp_dst_path, new_path)
                logging.info('Book converted & loaded: "{oldfile}" to "{newfile}"'.format(oldfile = filename, newfile = new_filename))
                bs.books_load()
        else: # 不接受其他檔案
            file_not_acceptable()
    # threading.Thread(target = index_update_necessary).start() # reindex
    index_update_necessary() # reindex
    bottle.redirect('/bookshelf.html') # to bookshelf
    # try: redirect(new_bid)
    # except NameError: pass

@app.put('/func/appended.head')
def appended_head_put():
    '''更新 append head'''
    bid = bottle.request.forms.getunicode('bid')
    appended_head = bottle.request.forms.getunicode('appended_head')
    book = bs.get(bid)
    try:
        book.toybox.data_entity_update('appended.head',
                data = appended_head)
    except IMNotExist:
        book.toybox.data_create('appended.head',
                data = appended_head, toy_type = 'text')
    logging.info('Book appended.head modified: "{bid}"'.format(bid = bid))

@app.put('/func/site-info')
def site_info_put():
    '''更新 site_name'''
    site_name = bottle.request.forms.getunicode('site_name')
    CONFIG['SITEINFO']['site_name'] = site_name
    # site_info_url = bottle.request.forms.getunicode('site_info_url')
    # CONFIG['SITEINFO']['site_info_url'] = site_info_url
    with open(CONFIG_FILE, mode = 'w', encoding = 'utf-8') as f:
        CONFIG.write(f)
    logging.info('Site info has modified.')

@app.post('/func/index-update')
def index_update():
    '''執行索引更新'''
    sync = bottle.request.query.sync # 強制指定同步模式？
    if sync:
        index_update_necessary()
    else:
        try:
            threading.Thread(target = index_update_necessary).start()
        except RuntimeError as e: # 在不允許多線程的機器上
            pass

@app.post('/func/shutdown')
def shutdown():
    '''關閉伺服器，結束程式'''
    _exitprogram()
    os._exit(0)

@app.route('/data-search.html')
@lonote.translate.renew
def data_search():
    '''進行搜索，回傳無格式結果頁'''
    key = bottle.request.query.key
    page = bottle.request.query.page
    if page: pagenum = int(page)
    else: pagenum = 1;
    return index_search(key, pagenum)

# ====================================
# 啟動伺服器
# ====================================

class LoserverError(Exception): pass

def assert_server_noexists(port):
    '''確認當前 server 不存在於指定端口，否則跳出 Exception'''
    try:
        UR.urlopen('http://localhost:{}'.format(port))
        raise LoserverError('server already exists.')
    except urllib.error.URLError:
        pass

def assert_server_exists(port):
    '''確認當前 server 存在於指定端口，否則跳出 Exception'''
    try:
        UR.urlopen('http://localhost:{}'.format(port))
    except urllib.error.URLError:
        raise LoserverError('server not exists.')

def _config_init():
    '''替 global CONFIG 初始化'''
    global CONFIG_FILE, CONFIG # CONFIG
    CONFIG_FILE = os.path.join(CFG_LONOTE_DIR, 'configure.ini')
    CONFIG = configparser.ConfigParser()
    # 數值初始化
    CONFIG['SITEINFO'] = collections.OrderedDict((
            ('site_name', ''),
            ('site_info_url', ''),
            ))
    CONFIG['AUTHCODE'] = collections.OrderedDict((
            ('full_access_code', ''),
            ('viewable_code', ''),
            ('code_hint', ''),
            ))
    # 拼入使用者設定的數值
    try_encoding = ['utf-8', 'utf-8-sig', None]
    for encoding in try_encoding: # 測試編碼
        try:
            CONFIG.read(CONFIG_FILE, encoding = encoding)
            break
        except: pass
    # 將完整的全新設定值回存
    with open(CONFIG_FILE, mode = 'w', encoding = 'utf-8') as f:
        CONFIG.write(f)

def _logging_init():
    '''替 logging 系統初始化'''
    global LOGGING_FILE
    LOGGING_FILE = os.path.join(CFG_LONOTE_DIR, 'logging.txt')
    logging.Formatter.converter = time.gmtime
    logging.basicConfig(filename = LOGGING_FILE,
            datefmt = '%Y-%m-%d %H:%M:%S +0000',
            format = '[%(asctime)s] %(levelname)s: %(message)s',
            level = 'NOTSET')

def _recent_edit_init():
    '''替 recent list 系統初始化'''
    global RECENT_EDITED_FILE # RECENT EDITED
    RECENT_EDITED_FILE = os.path.join(CFG_LONOTE_DIR,
            'recent_edited.pickle')
    global RECENT_EDITED
    try:
        with open(RECENT_EDITED_FILE, 'rb') as f:
            RECENT_EDITED = pickle.load(f)
    except: RECENT_EDITED = collections.deque(maxlen = 10)

def _custom_static_init():
    '''如果 cfgdir 下的 custom 資料夾不存在，新建之'''
    from_dir = os.path.join(STATIC_DIR, 'custom')
    to_dir = os.path.join(CFG_LONOTE_DIR, 'static', 'custom')
    try: shutil.copytree(from_dir, to_dir)
    except: pass

def init_paths(cfgdir, serdir):
    '''透過 cfgdir 與 serdir 設定各種路徑，並確保路徑存在

        cfgdir = 使用者資料與設定資料夾，可以讀寫使用者專有資料
        serdir = 伺服器資料夾，內含模版與靜態檔案等。
    '''
    # cfgdir
    global CFG_DIR, CFG_LONOTE_DIR
    CFG_DIR = cfgdir
    CFG_LONOTE_DIR = os.path.join(cfgdir, 'lonote')
    bookdir = os.path.join(CFG_LONOTE_DIR, 'book')
    global bs
    bs = Bookshelf(bookdir)
    global INDEX_DIR
    INDEX_DIR = os.path.join(CFG_LONOTE_DIR, 'index')
    global USER_STATIC_DIR
    USER_STATIC_DIR = os.path.join(CFG_LONOTE_DIR, 'static')
    for dir in (bookdir, INDEX_DIR, USER_STATIC_DIR): # 不存在則新建
        try: os.makedirs(dir, exist_ok = True)
        except OSError: pass

    _logging_init()
    _recent_edit_init()
    _config_init()

    # serdir
    static_dir = os.path.join(serdir, 'static')
    template_dir = os.path.join(serdir, 'template')

    global STATIC_DIR
    STATIC_DIR = static_dir
    global TEMPLATE_LOOKUP
    TEMPLATE_LOOKUP = mako.lookup.TemplateLookup(
            directories = [template_dir],
            input_encoding = 'utf-8')

    _custom_static_init()

def _exitprogram():
    bs.books_close()
    logging.info('Server shutdown')
    print('\nServer shutdown! Goodbye!')

@functools.lru_cache()
def get_wsgi(cfgdir, serdir, compress = False):
    '''取得最終用來執行的 wsgi 應用程式，並完成一切初始化工作

        cfgdir   = 放置使用者資料的 configure dir
        serdir   = 放置伺服器模板與靜態檔案的 server dir
        compress = 是否使用內建網頁壓縮？
                   如果使用的伺服器支援網頁壓縮，也可以讓伺服器來做

        return  => WSGI 應用程式
        '''
    init_paths(cfgdir, serdir)   # 設定路徑
    if compress:
        compressor_plugin = CompressorPlugin(compress_level = 6)
        app.install(compressor_plugin)

    real_wsgi = AuthMiddleware(app.wsgi)
    process_safe_wsgi = ProcessSafeMiddleware(real_wsgi)

    # 完成初始化
    bs.books_load()
    index_update_necessary()
    atexit.register(_exitprogram) # 註冊中止程式
    return process_safe_wsgi

def start_server(cfgdir, serdir,
        remote = False, port = 10204, debug = False, browser = ''):
    '''啟動內建伺服器'''

    try: # 測試伺服器是否已經在本地指定端口存在
        assert_server_noexists(port)
    except:
        print('... Server already exists!')
        if browser != None:
            print('Open browser...')
        open_browser(browser = browser, port = port)
        return

    # 啟動 browser，提前啟動的原因是因為伺服器啟動後我就沒控制權了……
    threading.Thread(
            target = open_browser,
            kwargs = {
                'browser': browser,
                'port': port}
            ).start()

    # ===== 正式啟動伺服器 =====
    if remote:
        compress = True
        host = '0.0.0.0'
    else:
        compress = False
        host = '127.0.0.1'
    if debug == True:
        quiet = False
    else:
        quiet = True

    final_wsgi = get_wsgi(cfgdir, serdir, compress = compress)

    print('Server starting...')
    logging.info('Server start')

    bottle.run(app = final_wsgi, host = host, port = port,
            debug = debug, quiet = quiet)

def open_browser(browser, port):
    '''打開 browser

        browser = None 表示不開啟
                = "" 表示用預設開啟
                = 其他字串 表示用指定的瀏覽器開啟
    '''
    open_target = 'http://localhost:{}'.format(port)
    loop = 0
    while True:
        time.sleep(0.2)
        try:  # 偵測port可否被讀取，有則打開瀏覽器
            assert_server_exists(port)
            if browser is None:
                pass
            else:
                if browser == '':
                    webbrowser.open(open_target)
                else:
                    try:
                        wb = webbrowser.get(browser)
                        wb.open(open_target)
                    except webbrowser.Error as e:
                        print(e)
            return
        except LoserverError:
            if loop > 600:
                break
            else:
                loop += 1
